# traversal.tcl --
#
#       Provides a conceptual tab group with bindings which enable traversal
#       between widgets to occur within and between tab groups. Within a tab
#       group, the arrow keys move the keyboard focus amongst widgets. The
#       The tab key moves keyboard focus between tab groups.
#
#       Please see the README for details, and either the test or example
#       files for examples.
#
#       Email: jjddss@my-deja.com
#
#       Copyright (c) 2000 Jay Schmidgall
#
#       except traversal::FocusOK, which is used by permission but is also
#
#       Copyright (c) 1994-1995 Sun Microsystems, Inc.
#
#       The author hereby grants permission to use, copy, modify, distribute,
#       and license this software and its documentation for any purpose, provided
#       that existing copyright notices are retained in all copies and that this
#       notice is included verbatim in any distributions. No written agreement,
#       license, or royalty fee is required for any of the authorized uses.
#       Modifications to this software may be copyrighted by their authors
#       and need not follow the licensing terms described here, provided that
#       the new terms are clearly indicated on the first page of each file where
#       they apply. 
#
#       IN NO EVENT SHALL THE AUTHOR OR DISTRIBUTORS BE LIABLE TO ANY PARTY
#       FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
#       ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
#       DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
#       POSSIBILITY OF SUCH DAMAGE.
#
#       THE AUTHOR AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
#       INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
#       FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
#       IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
#       NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
#       MODIFICATIONS.
#
#       GOVERNMENT USE: If you are acquiring this software on behalf of the
#       U.S. government, the Government shall have only "Restricted Rights"
#       in the software and related documentation as defined in the Federal 
#       Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
#       are acquiring the software on behalf of the Department of Defense, the
#       software shall be classified as "Commercial Computer Software" and the
#       Government shall have only "Restricted Rights" as defined in Clause
#       252.227-7013 (c) (1) of DFARs.  Notwithstanding the foregoing, the
#       author grants the U.S. Government and others acting in its behalf
#       permission to use and distribute the software in accordance with the
#       terms specified in this license.
# 

package provide traversal 0.9       ;# This uses Tcl namespaces

namespace eval traversal {
    variable grouping
    variable groupedby

    array set grouping {}
    array set groupedby {}

    # Note that truly, the only procs needed to be exported are makeTabGroup
    # and expand. However, exporting the others makes a testing framework
    # much easier.

    namespace export makeTabGroup traversePrev traverseNext traverseHome \
            traverseUp traverseLeft traverseDown traverseRight \
            traverseNextTab traversePrevTab expand
}


# traversal::makeTabGroup --
#
#       Establish the given widget as a tab group
#
# Arguments:
#       widget          The widget to make a tabgroup
#       container       Optional widget for which widget becomes a tab group.
#                       Defaults to widget parent if not specified.
#       method          Option method of traversal: standard, ordered or
#                       <name of command>
#       widgetlist      Optional list of widgets to make up the tabgroup
#                       for widget. Defaults to traversal::expand of widget.
#
# Results:
#       The given widget is made a tab group. Note that it is possible to
#       invoke multiple times for the same widget if you wish to change
#       the traversal method.
#
proc traversal::makeTabGroup { widget { container {} } { method standard } { widgetlist {} } } {
    variable grouping
    variable groupedby

    # Widgets of these types get handled specially

    set special Entry|Text|List

    # Establish the widgets in the tab group of widget if not passed in

    if { [ llength $widgetlist ] == 0 } {
        # No widgets in our tab grouping. Check if we're special and if we
        # are, we become our own tab group.

        if { [ regexp $special [ winfo class $widget ] ] == 1 } {
            set widgetlist $widget
        } else {
            # We weren't special :(. So expand ourself. If still nothing,
            # then we become our own tab group.

            set widgetlist [ traversal::expand $widget ]
            if { [ llength $widgetlist ] == 0 } {
                set widgetlist $widget
            }
        }
    }

    # Determine method of traversal
    #       standard    Up,down,left,right traversal
    #       order       Traversal done in order given
    #       other       Assumed to be a command to execute
    #                   for custom traversal

    if { [ string equal standard $method ] } {
        set lt { traversal::traverseLeft  %W }
        set up { traversal::traverseUp    %W }
        set rt { traversal::traverseRight %W }
        set dn { traversal::traverseDown  %W }
    } elseif { [ string equal order $method ] } {
        set lt { traversal::traversePrev %W }
        set up { traversal::traversePrev %W }
        set rt { traversal::traverseNext %W }
        set dn { traversal::traverseNext %W }
    } else {
        set lt [ list traversal::InvokeTakeFocus $method %W lt ]
        set up [ list traversal::InvokeTakeFocus $method %W up ]
        set rt [ list traversal::InvokeTakeFocus $method %W rt ]
        set dn [ list traversal::InvokeTakeFocus $method %W dn ]
    }

    # Establish the traversal bindings for the widget's group

    foreach tabstop $widgetlist {
        set groupedby($tabstop) $widget

        if { [ regexp $special [ winfo class $tabstop ] ] == 1 } {
            # Special widgets don't get all the bindings as they use the arrow
            # keys specially

            bind $tabstop <Control-Key-Tab>       { traversal::traverseNextTab %W ; break }
            bind $tabstop <Control-Shift-Key-Tab> { traversal::traversePrevTab %W ; break }
        } else {
            bind $tabstop <Key-Tab>       { traversal::traverseNextTab %W ; break }
            bind $tabstop <Shift-Key-Tab> { traversal::traversePrevTab %W ; break }
            bind $tabstop <Key-Up>        $up
            bind $tabstop <Key-Left>      $lt
            bind $tabstop <Key-Down>      $dn
            bind $tabstop <Key-Right>     $rt
            bind $tabstop <Key-Home>      { traversal::traverseHome %W ; break }
            bind $tabstop <Key-Find>      { traversal::traverseHome %W ; break }
        }
    }

    # Add widget to the container's grouping if it isn't there already

    if { [ string equal $container {} ] } {
        set container [ winfo parent $widget ]
    }

    if { [ info exists grouping($container) ] } {
        if { [ lsearch -exact $grouping($container) $widget ] == -1 } {
            lappend grouping($container) $widget
        }
    } else {
        lappend grouping($container) $widget
    }

    # Keep track of the widgets in this tabgroup

    set grouping($widget) $widgetlist

    # And our container groups the widget. Note this must be after the binding
    # loop above, as it is possible that widget is in the list of widgets. In
    # that case, this set would already have happened (incorrectly).
    
    set groupedby($widget) $container
    
    return
}


# traversal::expand --
#
#       Accumulate all widgets which are final descendants of the widgets in
#       the given list; that is, if any of the given widgets have children,
#       get them and add them to the list if they have no children. If they
#       do, get their children and add them to the list. This also applies
#       to slaves managed by the widget as well.
#
#       This is basically a convenience routine for making a tab group out of
#       a complex sequence of nested frames or such.
#
# Arguments:
#       widgetlist      The list of widgets
#       include         Optional flag whether to include widgets with
#                       children in the list
#
# Results:
#       A list containing all bottom-level children
#
proc traversal::expand { widgetlist { include 0 } } {
    set childlist {}
    foreach widget $widgetlist {
        # Flag no kids

        set nokids 1

        # Check for children

        set children [ winfo children $widget ]
        if { [ llength $children ] } {
            set nokids 0
            set morekids [ traversal::expand $children $include ]
            set childlist [ concat $childlist $morekids ]
        }

        # Check for packed slaves
        
        traversal::GetSlaves $widget pack $include nokids childlist

        # Check for gridded slaves
        
        traversal::GetSlaves $widget grid $include nokids childlist

        # Check for placed slaves

        traversal::GetSlaves $widget place $include nokids childlist

        # Check if we include this widget
        
        if { $include || $nokids } {
            lappend childlist $widget
        }
    }
    
    # Ensure uniqueness in childlist. This would probably be quicker to do
    # by simply setting the array element and returning array names. But
    # I want to maintain some semblance of the ordering.

    set returnlist {}
    foreach child $childlist {
        if { ! [ info exists kids($child) ] } {
            set kids($child) {}
            lappend returnlist $child
        }
    }
        
    return $returnlist
}


# traversal::GetSlaves --
#       Return the slaves of the given widget managed by the given method.
#       This routine is intended to only be called from traversal::expand.
#
# Arguments:
#       widget          The widget to check for slaves
#       method          By which method are the slaves managed
#       include         Indicate whether to include a widget with
#                       slaves in the list
#       nokidsName      Used for reference to nokids in callling proc
#       childlistName   Used for reference to childlist in calling proc
#
# Results:
#       Nokids will be cleared if the widget has slaves.
#       Childlist will have the slaves added to it.
#
proc traversal::GetSlaves { widget method include nokidsName childlistName } {
    upvar 1 $nokidsName nokids
    upvar 1 $childlistName childlist
    set slaves [ $method slaves $widget ]
    if { [ string equal grid $method ] } {
        # grid (currently) returns slaves in reverse order of
        # which they were gridded. So, we reverse the order.

        set reverse {}
        foreach temp $slaves {
            set reverse [ concat $temp $reverse ]
        }
        set slaves $reverse
    }                    
    if { [ llength $slaves ] } {
        set nokids 0
        set moreslaves [ traversal::expand $slaves $include ]
        set childlist [ concat $childlist $moreslaves ]
    }
    return
}


# traversal::Compare --
#
#       Compare the two elements based on their subelements
#       Each element is assumed to be a proper Tcl list of
#       three subelements, a widget followed by two numbers.
#
# Arguments:
#       first       The first element to compare.
#       second      The second element to compare.
#
# Results:
#       -1, 0 or 1 depending on the numbers in each element.
#
proc traversal::Compare { first second } {

    # Set indices to compare

    set idx1 2
    set idx2 1
    
    # Get the first elements to compare and do so

    set f [ lindex $first $idx1 ]
    set s [ lindex $second $idx1 ]
    if { $f > $s } {
        return 1
    }
    if { $f < $s } {
        return -1
    }

    # First elements were equal so now get the second elements
    # and compare them

    set f [ lindex $first $idx2 ]
    set s [ lindex $second $idx2 ]
    if { $f > $s } {
        return 1
    }
    if { $f < $s } {
        return -1
    }

    # Equality

    return 0
}


# traversal::FocusOK --
#
#       This procedure is invoked to decide whether or not to focus on
#       a given window.  It returns 1 if it's OK to focus on the window,
#       0 if it's not OK.  The code first checks whether the window is
#       viewable.  If not, then it never focuses on the window.  Then it
#       checks the -takefocus option for the window and uses it if it's
#       set.  If there's no -takefocus option, the procedure checks to
#       see if (a) the widget isn't disabled, and (b) it has some key
#       bindings.  If all of these are true, then 1 is returned.
#
#       This routine is a slightly modified version of tkFocusOk,
#       Copyright (c) 1994-1995 Sun Microsystems, Inc., and
#       used by permission.
#
# Arguments:
#       w 		Name of a widget.
#
# Results
#       Answer 1 if it is Ok to focus on the widget, 0 otherwise.

proc traversal::FocusOK { w } {
    set code [ catch { $w cget -takefocus } value ]

    if { ( $code == 0 ) && [ string compare $value "" ] } {
        if { $value == 0 } {
            return 0
        } elseif { $value == 1 } {
            return [ winfo viewable $w ]
        } else {
            set code [ catch { uplevel #0 $value $w } value ]
            if { ( $code == 0 ) && [ string compare $value "" ] } {
                return $value
            }
        }
    }

    if { ! [ winfo viewable $w ] } {
        return 0
    }

    set code [ catch {$w cget -state} value ]
    if { ( $code == 0 ) && [ string equal $value "disabled" ] } {
        return 0
    }

    return [ regexp Key|Focus "[ bind $w ] [ bind [winfo class $w] ]" ]
}

# traversal::HorizontalSlice --
#
#       Find all the widgets of this tab group which fall
#       within the horizontal slice defined by upper, lower.
#
# Arguments:
#       upper           The upper boundary of the slice
#       lower           The lower boundary of the slice
#       tabgroup        The tab group from which to check
#                       for widgets in the slice
#       reverse         (Optional) indicates reverse order
#
# Results:
#       An ordered list (by x position) of widgets which are
#       contained in the horizontal slice.
#
proc traversal::HorizontalSlice { upper lower tabgroup {reverse 0}} {
    variable grouping

    set slice {}
    set slicelist {}
    foreach tabstop $grouping($tabgroup) {
        if { [ traversal::FocusOK $tabstop ] == 0 } {
            continue
        }

        set x [ winfo rootx $tabstop ]
        set y1 [ winfo rooty $tabstop ]
        set y2 [ expr { $y1 + [ winfo height $tabstop ] } ]

        if { ( $upper <= $y1 ) && ( $y2 <= $lower ) } {
            lappend slicelist [ list $tabstop $x $y1 ]
        }
    }
        
    # I know, I could just do decreasing sort with -decreasing; however, I
    # need to be able to compare based on the other element if we have an
    # equal element, and -index only takes one index. Further, if by some
    # chance two elements are equal, odds are they'd show up in the same
    # relative position whether increasing or decreasing, and I need a
    # true reversal which may not happen in that case

    set slicelist [ lsort -command traversal::Compare -increasing $slicelist ]

    foreach tabstopinfo $slicelist {
        foreach { tabstop x unused } $tabstopinfo {
            if { $reverse } {
                set slice [ concat $tabstop $slice ]
            } else {
                lappend slice $tabstop
            }
        }
    }

    return $slice
}


# traversal::VerticalSlice- --
#
#       Find all the widgets of this tab group which fall
#       within the vertical slice defined by left, right.
#
# Arguments:
#       left            The leftmost boundary of the slice
#       right           The rightmost boundary of the slice
#       tabgroup        The tab group from which to check
#                       for widgets in the slice
#       reverse         (Optional) indicates reverse order
#
# Results:
#       An ordered list (by y position) of widgets which are
#       contained in the vertical slice.
#
proc traversal::VerticalSlice { left right tabgroup { reverse 0 } } {
    variable grouping

    set slice {}
    set slicelist {}
    foreach tabstop $grouping($tabgroup) {
        if { [ traversal::FocusOK $tabstop ] == 0 } { 
            continue
        }

        set x1 [ winfo rootx $tabstop ]
        set y [ winfo rooty $tabstop ]
        set x2 [ expr { $x1 + [ winfo width $tabstop ] } ]

        if { ( $left <= $x1 ) && ( $x2 <= $right ) } {
            lappend slicelist [ list $tabstop $y $x1 ]
        }
    }

    # I know, I could just do decreasing sort with -decreasing; however, I
    # need to be able to compare based on the other element if we have an
    # equal element, and -index only takes one index. Further, if by some
    # chance two elements are equal, odds are they'd show up in the same
    # relative position whether increasing or decreasing, and I need a
    # true reversal which may not happen in that case

    set slicelist [ lsort -command traversal::Compare -increasing $slicelist ]

    foreach tabstopinfo $slicelist {
        foreach { tabstop y unused } $tabstopinfo {
            if { $reverse } {
                set slice [ concat $tabstop $slice ]
            } else {
                lappend slice $tabstop
            }
        }
    }

    return $slice
}


# traversal::InvokeTakeFocus --
#
#       Use the value of the given widget's takeFocus configuration value
#       as a command to move the focus in the given direction.
#
# Arguments:
#       command     The command to execute
#       widget      The widget from which to move
#       direction   The direction in which to move
#
# Results:
#       The keyboard focus should now be on the appropriate widget in the
#       same tab group as a result of invoking user code.
#
proc traversal::InvokeTakeFocus { command widget direction } {
    catch { uplevel #0 $command $widget $direction }

    return [ focus -lastfor $widget ]
}


# traversal::TraverseDir --
#
#       Move focus in the given direction
#       from the given widget
#
# Arguments:
#       widget      The widget from which to move
#       direction   The direction in which to move
#
# Results:
#       The keyboard focus should now be on the appropriate widget in the
#       same tab group.
#
proc traversal::TraverseDir { widget direction } {
    variable groupedby
    
    set tabgroup $groupedby($widget)
    
    # Get widget boundaries

    set left [ winfo rootx $widget ]
    set upper [ winfo rooty $widget ]
    set right [ expr { $left + [ winfo width $widget ] } ]
    set lower [ expr { $upper + [ winfo height $widget ] } ]

    # This is pretty much the heart of the matter. We make 3 passes.
    # First pass: get a slice which exactly contains the widget.
    # If there's a widget in there, move focus to that widget.
    # If not, look all around the widget for another widget in
    # that area. Otherwise, we basically end up wrapping to the
    # first widget that we found in the previous pass.

    set looping 0
    while { 1 } {
        switch -- $direction {
            up {
                set slice [ traversal::VerticalSlice $left $right $tabgroup 1 ]
            }
            dn {
                set slice [ traversal::VerticalSlice $left $right $tabgroup ]
            }
            lt {
                set slice [ traversal::HorizontalSlice $upper $lower $tabgroup 1 ]
            }
            rt {
                set slice [ traversal::HorizontalSlice $upper $lower $tabgroup ]
            }
        }

        if { [ llength $slice ] == 0 } {
            set slice $widget
        }

        set next [ lsearch $slice $widget ]
        incr next

        if { $next >= [ llength $slice ] } {
            if { $looping > 1 } {
                focus [ lindex $slice 0 ]
                break
            }
            incr looping

            if { $looping == 1 } {
                # Reset boundaries to be all inclusive

                set left 0
                set upper 0
                set right [ winfo screenwidth $widget ]
                set lower [ winfo screenheight $widget ]
            }
        } else {
            focus [ lindex $slice $next ]
            break
        }
    }
    set looping 0
    return [ focus -lastfor $widget ]
}


# traversal::traverseUp --
#
#       Move focus to the widget above the given widget,
#       wrapping if necessary.
#
# Arguments:
#       widget      Reference widget from which to move up
#
# Results:
#       The keyboard focus should now be on the widget above
#       of the given widget.
#
proc traversal::traverseUp { widget } {
    return [ traversal::TraverseDir $widget up ]
}


# traversal::traverseLeft --
#
#       Move focus to the widget on the left of the given widget,
#       wrapping if necessary.
#
# Arguments:
#       widget      Reference widget from which to move left
#
# Results:
#       The keyboard focus should now be on the widget to the left
#       of the given widget.
#
proc traversal::traverseLeft { widget } {
    return [ traversal::TraverseDir $widget lt ]
}


# traversal::traverseDown --
#
#       Move focus to the widget below the given widget,
#       wrapping if necessary.
#
# Arguments:
#       widget      Reference widget from which to move down
#
# Results:
#       The keyboard focus should now be on the widget below
#       of the given widget.
#
proc traversal::traverseDown { widget } {
    return [ traversal::TraverseDir $widget dn ]
}


# traversal::traverseRight --
#
#       Move focus to the widget on the right of the given widget,
#       wrapping if necessary.
#
# Arguments:
#       widget      Reference widget from which to move right
#
# Results:
#       The keyboard focus should now be on the widget to the right
#       of the given widget.
#
proc traversal::traverseRight { widget } {
    return [ traversal::TraverseDir $widget rt ]
}


# traversal::traverseHome --
#
#       Move focus to the first widget in the tab group.
#
# Arguments:
#       widget      Reference widget from which to move left
#
# Results:
#       The keyboard focus should now be on the widget in the first position
#       of the tab group order.
#
proc traversal::traverseHome { widget } {
    variable grouping
    variable groupedby
    
    set tabgroup $groupedby($widget)
    set tabstop [ lindex $grouping($tabgroup) 0 ]
    if { [ traversal::FocusOK $tabstop ] } {
        focus $tabstop
    } else {
        traversal::traverseNext $tabstop
    }

    return [ focus -lastfor $widget ]
}


# traversal::traverseNext --
#
#       Move focus to the next tab stop.
#
# Arguments:
#       widget      Reference widget from which to move to the next
#
# Results:
#       The keyboard focus should now be on the next widget in the tab group
#       order.
#
proc traversal::traverseNext { widget } {
    variable grouping
    variable groupedby
    
    set tabgroup $groupedby($widget)
    set currentstop [ lsearch -exact $grouping($tabgroup) $widget ]
    set next [ expr { $currentstop + 1 } ]
    if { $next >= [ llength $grouping($tabgroup) ] } {
        set next 0
    }

    while { $next != $currentstop } {
        set tabstop [ lindex $grouping($tabgroup) $next ]
        if { [ traversal::FocusOK $tabstop ] } {
            focus $tabstop
            break
        }
        incr next
        if { $next >= [ llength $grouping($tabgroup) ] } {
            set next 0
        }
    }

    return [ focus -lastfor $widget ]
}


# traversal::traversePrev --
#
#       Move focus to the previous tab stop.
#
# Arguments:
#       widget      Reference widget from which to move to the previous
#
# Results:
#       The keyboard focus should now be on the previous widget in the tab
#       group order.
#
proc traversal::traversePrev { widget } {
    variable grouping
    variable groupedby
    
    set tabgroup $groupedby($widget)
    set currentstop [ lsearch -exact $grouping($tabgroup) $widget ]
    set prev [ expr { $currentstop - 1 } ]
    if { $prev < 0 } {
        incr prev [ llength $grouping($tabgroup) ]
    }

    while { $prev != $currentstop } {
        set tabstop [ lindex $grouping($tabgroup) $prev ]
        if { [ traversal::FocusOK $tabstop ] } {
            focus $tabstop
            break
        }
        incr prev -1
        if { $prev < 0 } {
            incr prev [ llength $grouping($tabgroup) ]
        }
    }

    return [ focus -lastfor $widget ]
}


# traversal::traverseNextTab --
#
#       Move focus to the next tab group.
#
# Arguments:
#       widget      Reference widget from which to move to the next tab stop
#
# Results:
#       The keyboard focus should now be on the home widget in the next tab
#       group.
#
proc traversal::traverseNextTab { widget } {
    variable grouping
    variable groupedby
    
    set tabgroup $groupedby($widget)
    set container $groupedby($tabgroup)
    set currentgroup [ lsearch -exact $grouping($container) $tabgroup ]
    set next [ expr { $currentgroup + 1 } ]
    if { $next >= [ llength $grouping($container) ] } { 
        set next 0
    }
    
    # Advance to the next group with a focusable widget. If no such group
    # exists, then leave focus where it is at

    while { $next != $currentgroup } {
        set nextgroup [ lindex $grouping($container) $next ]
        traversal::traverseHome [ lindex $grouping($nextgroup) 0 ]
        set lastfocus [ focus -lastfor $widget ]
        if { [ string equal $widget $lastfocus ] == 0 } {
            break
        }
        incr next
        if { $next >= [ llength $grouping($container) ] } { 
            set next 0
        }
    }

    return [ focus -lastfor $widget ]
}


# traversal::traversePrevTab --
#
#       Move focus to the previous tab group.
#
# Arguments:
#       widget      Reference widget from which to move to the previous tab
#                   group
#
# Results:
#       The keyboard focus should now be on the home widget in the previous
#       tab group.
#
proc traversal::traversePrevTab { widget } {
    variable grouping
    variable groupedby
    
    set tabgroup $groupedby($widget)
    set container $groupedby($tabgroup)
    set currentgroup [ lsearch -exact $grouping($container) $tabgroup ]
    set currentgroup [ lsearch -exact $grouping($container) $tabgroup ]
    set prev [ expr { $currentgroup - 1 } ]
    if { $prev < 0 } { 
        incr prev [ llength $grouping($container) ]
    }

    # Move focus to the previous group with a focusable widget. If no such
    # group exists, then leave focus where it is at

    while { $prev != $currentgroup } {
        set prevgroup [ lindex $grouping($container) $prev ]
        traversal::traverseHome  [ lindex $grouping($prevgroup) 0 ]
        set lastfocus [ focus -lastfor $widget ]
        if { [ string equal $widget $lastfocus ] == 0 } {
            break
        }
        incr prev -1
        if { $prev < 0 } { 
            incr prev [ llength $grouping($container) ]
        }
    }
   
    return [ focus -lastfor $widget ]
}
